import asyncio
from typing import Any, Optional, Union

from patio.broker.abc import AbstractBroker, TimeoutType
from patio.registry import TaskFunctionType


class MemoryBroker(AbstractBroker):
    async def call(
        self,
        func: Union[str, TaskFunctionType],
        *args: Any,
        timeout: Optional[TimeoutType] = 86400,
        **kwargs: Any,
    ) -> Any:
        return await asyncio.wait_for(
            self.executor.execute(func, *args, **kwargs),
            timeout=timeout,
        )


__all__ = "MemoryBroker",



from __future__ import annotations

import hashlib
import threading
from dataclasses import dataclass
from enum import IntEnum, unique
from io import BytesIO
from random import getrandbits
from struct import Struct
from typing import Any, Optional, Tuple

from patio.broker.serializer import AbstractSerializer


@unique
class PacketTypes(IntEnum):
    AUTH_DIGEST = 0
    AUTH_REQUEST = 1
    AUTH_RESPONSE = 2
    AUTH_OK = 3
    REQUEST = 10
    RESPONSE = 20
    ERROR = 30


@dataclass
class Header:
    STRUCT = Struct("!bII")
    SIZE = STRUCT.size

    type: PacketTypes
    size: int
    serial: int

    def pack(self) -> bytes:
        return self.STRUCT.pack(self.type.value, self.size, self.serial)

    @classmethod
    def unpack(cls, data: bytes) -> Header:
        kind, size, serial = cls.STRUCT.unpack(data)
        return cls(type=PacketTypes(kind), size=size, serial=serial)


class Protocol:
    HEADER_STRUCT = Struct("!bII")
    MAX_SERIAL = 4294967295

    __slots__ = "__key", "serial", "lock", "serializer"

    def __init__(self, *, serializer: AbstractSerializer, key: bytes = b""):
        self.__key = key
        self.serial = 0
        self.lock = threading.Lock()
        self.serializer = serializer

    def get_serial(self) -> int:
        with self.lock:
            self.serial += 1

            if self.serial >= self.MAX_SERIAL:
                self.serial = 1

            return self.serial

    def digest(
        self, data: bytes, *, salt: Optional[bytes] = None
    ) -> Tuple[bytes, bytes]:
        if salt is None:
            salt = getrandbits(32).to_bytes(4, "big")
        return salt, hashlib.blake2s(data, key=self.__key, salt=salt).digest()

    def pack(
        self, payload: Any, packet_type: PacketTypes,
        *, serial: Optional[int] = None
    ) -> bytes:
        with BytesIO() as fp:
            fp.seek(Header.SIZE)
            fp.write(self.serializer.pack(payload))

            header = Header(
                type=packet_type,
                size=fp.tell() - self.HEADER_STRUCT.size,
                serial=serial or self.get_serial(),
            )

            fp.seek(0)
            fp.write(header.pack())

            return fp.getvalue()

    def authorize_request(self, token: bytes) -> bytes:
        salt, digest = self.digest(token)
        return self.pack((salt, digest, token), PacketTypes.AUTH_REQUEST)

    def authorize_check(self, payload: bytes) -> bool:
        salt, digest, token = self.serializer.unpack(payload)
        return self.digest(token, salt=salt) == (salt, digest)

    def pack_error(self, exception: Exception, serial: int) -> bytes:
        return self.pack(exception, PacketTypes.ERROR, serial=serial)


__all__ = ("PacketTypes", "Header", "Protocol")



from typing import Any, Awaitable, Callable

from patio.executor.base import AbstractExecutor
from patio.registry import T


class NullExecutor(AbstractExecutor):
    """
    Doesn't execute anything, serves as a stub to explicitly forbid calls
    to this registry.
    """

    async def setup(self) -> None:
        pass

    def submit(
        self, func: Callable[..., T], *args: Any, **kwargs: Any
    ) -> Awaitable[T]:
        raise RuntimeError("Null executor can't execute anything")

    async def shutdown(self) -> None:
        pass



from .abc import AbstractBroker, TimeoutType
from .memory import MemoryBroker
from .tcp.broker import TCPBrokerBase, TCPClientBroker, TCPServerBroker


__all__ = (
    "AbstractBroker",
    "MemoryBroker",
    "TCPBrokerBase",
    "TCPClientBroker",
    "TCPServerBroker",
    "TimeoutType",
)



import asyncio
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any, Awaitable, Coroutine, List, Optional, Set, Union

from patio.compat import Self
from patio.executor import AbstractExecutor
from patio.registry import TaskFunctionType


TimeoutType = Union[int, float]


class AbstractBroker(ABC):
    def __init__(self, executor: AbstractExecutor):
        self.executor = executor
        self.__tasks: Set[asyncio.Task] = set()

    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    def create_task(self, coro: Coroutine[Any, Any, Any]) -> asyncio.Task:
        task = self.loop.create_task(coro)
        self.__tasks.add(task)
        task.add_done_callback(self.__tasks.discard)
        return task

    async def setup(self) -> None:
        await self.executor.setup()

    async def close(self) -> None:
        tasks: List[Awaitable[Any]] = [self.executor.shutdown()]
        for task in tuple(self.__tasks):
            if task.done():
                continue
            task.cancel()
            tasks.append(task)
        await asyncio.gather(*tasks, return_exceptions=True)

    @abstractmethod
    async def call(
        self,
        func: Union[str, TaskFunctionType],
        *args: Any,
        timeout: Optional[TimeoutType] = None,
        **kwargs: Any,
    ) -> Any:
        ...

    async def __aenter__(self) -> Self:
        await self.setup()
        return self

    async def __aexit__(
        self, exc_type: Any, exc_val: Any, exc_tb: Any,
    ) -> None:
        await self.close()

    async def join(self) -> None:
        async def waiter() -> None:
            await self.loop.create_future()
        await self.create_task(waiter())


__all__ = "AbstractBroker", "TimeoutType", "TaskFunctionType"



from __future__ import annotations

import builtins
import inspect
import pickle
from abc import ABC, abstractmethod
from io import BytesIO
from typing import Any, Iterator


class AbstractSerializer(ABC):
    @abstractmethod
    def pack(self, payload: Any) -> bytes:
        ...

    @abstractmethod
    def unpack(self, payload: bytes) -> Any:
        ...


class PickleSerializer(AbstractSerializer):
    def pack(self, payload: Any) -> bytes:
        return pickle.dumps(payload, protocol=pickle.HIGHEST_PROTOCOL)

    def unpack(self, payload: bytes) -> Any:
        return pickle.loads(payload)


def _builtins_exceptions() -> Iterator[str]:
    for item in dir(builtins):
        obj = getattr(builtins, item)
        if not inspect.isclass(obj):
            continue
        if issubclass(obj, BaseException):
            yield f"builtins.{obj.__name__}"


class RestrictedUnpickler(pickle.Unpickler):
    SAFE_EXCLUDES = frozenset({
        "builtins.complex",
        "builtins.set",
        "builtins.frozenset",
        *_builtins_exceptions(),
    })

    UNSAFE_MODULES = frozenset(
        {"builtins", "os", "sys", "posix", "_winapi"},
    )

    def find_class(self, module: str, name: str) -> Any:
        if module not in self.UNSAFE_MODULES:
            return super().find_class(module, name)

        if f"{module}.{name}" in self.SAFE_EXCLUDES:
            return super().find_class(module, name)

        raise pickle.UnpicklingError(f"The '{module}.{name}' is forbidden")


class RestrictedPickleSerializer(PickleSerializer):
    def unpack(self, payload: bytes) -> Any:
        with BytesIO(payload) as fp:
            return RestrictedUnpickler(fp).load()
